/* 
 *  play_ac3.c
 *
 *	Copyright (C) 2000 by Andy Lo A Foe 
 *
 *  play_ac3 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  play_ac3 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/asoundlib.h>
#include "ac3.h"
#include "bitstream.h"
#include "parse.h"

#define BLOCK_SIZE	6144
//#define DEBUG 1

int pcm_mode = SND_PCM_MODE_BLOCK;
int frame_count = 0;

static 
void decode_find_sync(bitstream_t *bs)
{
	uint_16 sync_word;
	uint_32 i = 0;

	sync_word = bitstream_get(bs,16);

	/* Make sure we sync'ed */
	while(1)
	{
		if(sync_word == 0x0b77)
			break;
		sync_word = sync_word * 2;
		sync_word |= bitstream_get(bs,1);
		i++;
	}
	printf("(sync) %ld bits skipped to synchronize\n",i);
	printf("(sync) begin frame %ld\n",frame_count);
	frame_count++;

	bs->total_bits_read = 16;
	crc_init();
}



int main(int argc, char *argv[])
{
	snd_pcm_t *handle;
	snd_pcm_channel_info_t info;
	snd_pcm_digital_t	digital;
	snd_pcm_channel_setup_t setup;
	snd_pcm_channel_params_t params;
	bitstream_t *bs, store;
	int val,need_swab = 1;
	int old_rate = -1;
	syncinfo_t si;
	long pos;
	int err;
	char tmp[BLOCK_SIZE];
	char buf[BLOCK_SIZE];
	unsigned short *sbuf = (unsigned short *)buf;
	int c=0, d, e;
	short left=0;
	FILE *bla, *bla2;

	if (argc < 2) {
			printf("usage: ac3 file\n");
			return 1;
	}

					
	if ((err = snd_pcm_open(&handle, 0, 2, SND_PCM_OPEN_PLAYBACK)) < 0)
	{
		fprintf(stderr, "open: %s\n", snd_strerror(err));
		return 1;
	}
	memset(&info, 0, sizeof(info));
	if ((err = snd_pcm_channel_info(handle, &info)) < 0) {
		fprintf(stderr, "info: %s\n", snd_strerror(err));
		snd_pcm_close(handle);
		return 1;
	}
#ifdef DEBUG	
	printf("%s\n", info.subname);
	printf("channel = %d\n", info.channel);
	printf("formats = %d\n", info.formats);
	printf("flags = %d\n", info.flags);
	printf("samplerate = %d - %d\n", info.min_rate, info.max_rate);
	printf("mode = %d\n", info.mode);
	printf("fifo_size = %d\n", info.fifo_size);
	printf("buffer_size = %d\n", info.buffer_size);
#endif
	memset(&setup, 0, sizeof(setup));
	memset(&params, 0, sizeof(params));
	params.mode = pcm_mode;
	params.channel = SND_PCM_CHANNEL_PLAYBACK;
	params.start_mode = SND_PCM_START_FULL;
	params.stop_mode = SND_PCM_STOP_STOP;
	params.buf.block.frag_size = BLOCK_SIZE;
	params.buf.block.frags_max = 12;
	params.buf.block.frags_min = 3;
	params.digital.dig_valid = 1;
	params.digital.dig_status[0] = SND_PCM_DIG0_NONAUDIO;
	params.digital.dig_status[3] = SND_PCM_DIG3_CON_FS_48000;
	params.format.format = SND_PCM_SFMT_S16_LE;
	params.format.rate = 48000;
	params.format.interleave = 1;
	params.format.voices = 2;
	snd_pcm_channel_flush(handle, SND_PCM_CHANNEL_PLAYBACK);
			
	if ((err = snd_pcm_channel_params(handle, &params)) < 0) {
		fprintf(stderr, "params: %s\n", snd_strerror(err));
		snd_pcm_close(handle);
		return 3;
	}	
#if 1
	setup.digital.dig_valid = 1;
	setup.digital.dig_status[0] = SND_PCM_DIG0_NONAUDIO;
	setup.digital.dig_status[3]	= SND_PCM_DIG3_CON_FS_48000;
#endif
	setup.mode = pcm_mode;
	setup.channel = SND_PCM_CHANNEL_PLAYBACK;
	if ((err = snd_pcm_channel_prepare(handle, SND_PCM_CHANNEL_PLAYBACK)) < 0) {
		fprintf(stderr, "prepare: %s\n", snd_strerror(err));
		return 2;
	}
	if ((err = snd_pcm_channel_setup(handle, &setup)) < 0) {
		fprintf(stderr, "setup: %s\n", snd_strerror(err));
		snd_pcm_close(handle);
		return 2;
	}
	if (strcmp(argv[1], "-") == 0) {
			bla = stdin;
	} else		
		bla = fopen(argv[1], "r");
	if (!bla) {
			fprintf(stderr, "Invalid file\n");
			snd_pcm_close(handle);
			return 1;
	}		
#ifdef DEBUG
	bla2 = fopen("/tmp/ac3dump", "w");
#endif	
	
	bs = bitstream_open(bla);
	if (!bs) {
			printf("Error opening bitstream\n");
			exit(1);
	}
	memset(buf, 0, sizeof(buf));

	// Send some trailing zeros to avoid misdetecting the 
	// first frame as PCM (works for DTT2500 at least)
	for (c=0; c < 5; c++) {
		snd_pcm_write(handle, buf, BLOCK_SIZE);
	}

	// Read the whole file
	while (!feof(bla)) { 
		sbuf[0] = 0xf872; // Syncword
		sbuf[1] = 0x4e1f; // ........
		sbuf[2] = 0x0001; // AC-3 data
		// Fill in 3 later
		
		// Read AC3 syncwrod and crc field
		left = 32; // 2 x 16-bit fields
		c = 8;
		while (left > 0) {
				buf[c++] = bitstream_get(bs,8);
				left -= 8;
		}
		
		// Determine bitrate and frame_size;
		val = bitstream_get(bs,8);
		si.fscod = (val & 0x3f) >> 6;
		si.frmsizecod = (val & 0x3f);
		si.bit_rate = frmsizecod_tbl[si.frmsizecod].bit_rate;
		si.frame_size = frmsizecod_tbl[si.frmsizecod].frm_size[si.fscod];
		sbuf[3] = si.frame_size * 16;
		if (old_rate != si.bit_rate) {
			printf("Stream bitrate: %4d, frame_size: %4d\n", si.bit_rate, si.frame_size);
			old_rate = si.bit_rate;
		}	
		// Push the previous 8 bits into the frame
		buf[c++] = val;

		// Read in the rest of the AC3 frame
		left = si.frame_size * 16 - 40; // 16:sync, 16:CRC, 8:bitrate/frsize
		while (left > 0) {
				buf[c++] = bitstream_get(bs,8);
				left -= 8;
		}

		// extract_ac3 seems to write swabbed data
		swab(&buf[8], &buf[8], si.frame_size * 2);
		
#ifdef DEBUG
		if (bla2) 
			fwrite(buf, BLOCK_SIZE, 1, bla2);
#endif

		// Write out the frame to S/PDIF
		if ((snd_pcm_write(handle, buf, BLOCK_SIZE)) < 0) {
			printf("error writing\n");
			snd_pcm_close(handle);
			if (bla != stdin)
					fclose(bla);
			return 3;
		}
	}

	// Cleanup
	snd_pcm_close(handle);
	if (bla != stdin)
			fclose(bla);
#ifdef DEBUG
	if (bla2)
		fclose(bla2);
#endif	
	return 0;
}

